@@ -3,6 +3,7 @@
 
 #include <stddef.h>
 #include <stdint.h>
+#include <xf86drm.h>
 #include <xf86drmMode.h>
 
 extern uint32_t liftoff_mock_drm_crtc_id;
@@ -22,8 +23,11 @@ liftoff_mock_drm_open(void);
 uint32_t
 liftoff_mock_drm_create_fb(struct liftoff_layer *layer);
 
+void
+liftoff_mock_drm_set_fb_info(const drmModeFB2 *fb_info);
+
 struct liftoff_mock_plane *
-liftoff_mock_drm_create_plane(int type);
+liftoff_mock_drm_create_plane(uint64_t type);
 
 struct liftoff_mock_plane *
 liftoff_mock_drm_get_plane(uint32_t id);
@@ -33,9 +37,16 @@ liftoff_mock_plane_add_compatible_layer(struct liftoff_mock_plane *plane,
 					struct liftoff_layer *layer);
 struct liftoff_layer *
 liftoff_mock_plane_get_layer(struct liftoff_mock_plane *plane);
+uint32_t
+liftoff_mock_plane_get_id(struct liftoff_mock_plane *plane);
 
 uint32_t
 liftoff_mock_plane_add_property(struct liftoff_mock_plane *plane,
-				const drmModePropertyRes *prop);
+				const drmModePropertyRes *prop,
+				uint64_t value);
+void
+liftoff_mock_plane_add_in_formats(struct liftoff_mock_plane *plane,
+				  const struct drm_format_modifier_blob *data,
+				  size_t size);
 
 #endif
